@@ -15,22 +15,23 @@ SEXP R_matpow(SEXP x, SEXP k)
 	/*-Wall */ return R_NilValue;
     }
     else {
-	SEXP dims = getAttrib(x, R_DimSymbol), z, x_;
+	SEXP dims = getAttrib(x, R_DimSymbol);
 	int n = INTEGER(dims)[0],
-	    ktmp = INTEGER(k)[0]; /* need copy, as it is altered in matpow() */
+	    k_ = INTEGER(k)[0]; /* need copy, as it is altered in matpow() */
 
 	if (n != INTEGER(dims)[1])
 	    error(_("non-square matrix"));
 	if (n == 0)
 	    return(allocMatrix(REALSXP, 0, 0));
 
-	PROTECT(x_= duplicate(x));
+	SEXP x_ = duplicate(x);	PROTECT_INDEX xpi;
+	PROTECT_WITH_INDEX(x_, &xpi);
 	if (!isReal(x)) /* coercion to numeric */
-	    x_= coerceVector(x_, REALSXP);
-	PROTECT(z = allocMatrix(REALSXP, n, n));
+	    REPROTECT(x_ = coerceVector(x_, REALSXP), xpi);
+	SEXP z = PROTECT(allocMatrix(REALSXP, n, n));
 	setAttrib(z, R_DimNamesSymbol,
 		  getAttrib(x, R_DimNamesSymbol));
-	matpow(REAL(x_), n, ktmp, REAL(z));
+	matpow(REAL(x_), n, k_, REAL(z));
 
 	UNPROTECT(2);
 	return z;
@@ -55,7 +56,7 @@ void matpow(double *x, int n, int k, double *z)
     else { /* k >= 1 */
 	static const char *transa = "N";
 	static const double one = 1.0, zero = 0.0;
-	int nSqr = n * n;
+	size_t nSqr = n * ((size_t) n);
 	double /* temporary matrix */
 	    *tmp  = (double *) R_alloc(nSqr, sizeof(double));
 
@@ -63,14 +64,14 @@ void matpow(double *x, int n, int k, double *z)
 	 * product left to make. That is, if k = 5, compute (x * x),
 	 * then ((x * x) * (x * x)) and finally ((x * x) * (x * x)) * x.
 	 */
-	Memcpy(z, x, (size_t) nSqr);
+	Memcpy(z, x, nSqr);
 
 	k--;
 	while (k > 0) {
 	    if (k & 1) {	/* z := z * x */
 		F77_CALL(dgemm)(transa, transa, &n, &n, &n, &one,
-				z, &n, x, &n, &zero, tmp, &n);
-		Memcpy(z, tmp, (size_t) nSqr);
+				z, &n, x, &n, &zero, tmp, &n FCONE FCONE);
+		Memcpy(z, tmp, nSqr);
 	    }
 	    if(k == 1)
 		break;
@@ -78,8 +79,8 @@ void matpow(double *x, int n, int k, double *z)
 
 	    /* x := x * x */
 	    F77_CALL(dgemm)(transa, transa, &n, &n, &n, &one,
-			    x, &n, x, &n, &zero, tmp, &n);
-	    Memcpy(x, tmp, (size_t) nSqr);
+			    x, &n, x, &n, &zero, tmp, &n FCONE FCONE);
+	    Memcpy(x, tmp, nSqr);
 	}
     }
 }
